@@ -74,7 +74,7 @@ accidental dependencies.
 
 (define *interactive* #false)  ;; be silent
 (define *include-dirs* '(".")) ;; now we can (import <libname>) and have them be autoloaded to current repl
-(define *owl-version* "0.2.1")
+(define *owl-version* "0.2.2")
 
 (import
    (owl intern)
@@ -360,7 +360,7 @@ Check out https://haltp.org/owl for more information.")
                               (thread 'repl
                                  (let ((state (make-variable '*state* empty)))
                                     ;; get basic io running
-                                    (start-base-threads)
+                                    (start-io-threads)
 
                                     ;; store initial state values
                                     (state 'call
